--[[
MultiTool Multi Boxing utilities inspired by smurfys MultiBoxer addon

Shouts to http://old.wowace.com/wiki/WelcomeHome_-_Your_first_Ace3_Addon

Extra special shouts to Allara and OrionShock for pointing me in the right direction RE: AceComm-3.0

Author: DigitalSorceress
Date: 2008/11/18
Version: 0.2.01
]]

-- Some initialization of our happy local vars
local myName = "MultiTool"
--local myVersion = GetAddonMetadata(myName, "Version")
local myVersion = "v0.2.01"

local totalBagSlots = 0
local totalFreeBagSlots = 0
local totalUsedBagSlots = 0
local totalFreeQuestLogSlots = 0

MultiTool = LibStub("AceAddon-3.0"):NewAddon("MultiTool", "AceConsole-3.0","AceComm-3.0","AceEvent-3.0","AceSerializer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("MultiTool")

-- ------------------------------------------------
-- CONFIGURATION STUFF
-- ------------------------------------------------
-- Holds our configuration options
local options = {
  name = "MultiTool",
  handler = MultiTool,
  type = 'group',
  args = {
    debugOptions = {
      type = "group",
      name = L["DEBUG_GROUP_NAME"],
      order = 0,
      args = {
        debugFlag = {
          type = "toggle",
          name = L["DEBUG_NAME"],
          desc = L["DEBUG_DESC"],
          order = 1,
          get = "getDebugFlag",
          set = "setDebugFlag"
        },
        blatherFlag = {
          type = "toggle",
          name = L["BLATHER_NAME"],
          desc = L["BLATHER_DESC"],
          order = 2,
          get = "getBlatherFlag",
          set = "setBlatherFlag"
        }
      }
    },
    taxiFlag = {
      type = "toggle",
      name = L["TAXI_NAME"],
      desc = L["TAXI_DESC"],
      order = 1,
      get = "getTaxiFlag",
      set = "setTaxiFlag"
    },
    questOptions = {
      type = "group",
      name = L["QUEST_GROUP_NAME"],
      order = 2,
      args = {
        gossipFlag = {
           type = "toggle",
           name = L["GOSSIP_NAME"],
           desc = L["GOSSIP_DESC"],
           get = "getGossipFlag",
           set = "setGossipFlag"
         },   
        rewardFlag = {
           type = "toggle",
           name = L["REWARD_CLONE_NAME"],
           desc = L["REWARD_CLONE_DESC"],
           get = "getRewardFlag",
           set = "setRewardFlag"
         }
       }
    },
    vendorOptions = {
      type = "group",
      name = L["VENDOR_GROUP_NAME"],
      order = 3,
      args = {
        repairFlag= {
          type = "toggle",
          name = L["AUTO_REPAIR_NAME"],
          desc = L["AUTO_REPAIR_DESC"],
          order = 1,
          get = "getRepairFlag",
          set = "setRepairFlag"
        },
        vendJunkFlag= {
          type = "toggle",
          name = L["VEND_JUNK_NAME"],
          desc = L["VEND_JUNK_DESC"],
          order = 2,
          get = "getVendJunkFlag",
          set = "setVendJunkFlag"
        }
      }
    },
    monitorBag = {
      type = "group",
      name = L["BAG_MONITOR_NAME"],
      order = 4,
      args = {
        bagFlag = {
          type = "toggle",
          name = L["BAG_FLAG_NAME"],
          desc = L["BAG_FLAG_DESC"],
          order = 1,
          get = "getBagFlag",
          set = "setBagFlag"
        },
        bagWarn = {
          type = "range",
          min = 0,
          max = 10,
          step = 1,
          bigStep = 1,
          name = L["BAG_WARN_NAME"],
          desc = L["BAG_WARN_DESC"],
          order = 2,
          get = "getBagWarn",
          set = "setBagWarn"
        }
      }
    }, 
    monitorQuestLog = {
      type = "group",
      name = L["QUEST_LOG_MONITOR_NAME"],
      order = 5,
      args = {
        questLogFlag = {
          type = "toggle",
          name = L["QUEST_LOG_FLAG_NAME"],
          desc = L["QUEST_LOG_FLAG_DESC"],
          order = 1,
          get = "getQuestLogFlag",
          set = "setQuestLogFlag"
        },
        questLogWarn = {
          type = "range",
          min = 0,
          max = 10,
          step = 1,
          bigStep = 1,
          name = L["QUEST_LOG_WARN_NAME"],
          desc = L["QUEST_LOG_WARN_DESC"],
          order = 2,
          get = "getQuestLogWarn",
          set = "setQuestLogWarn"
        }
      }
    }, 
   duelFlag = {
      type = "toggle",
      name = L["DUELING_NAME"],
      desc = L["DUELING_DESC"],
      order = 6,
      get = "getDuelFlag",
      set = "setDuelFlag"
    }
  }
}


--
-- Set the default profile if they don't have one
--
local defaults = {
  profile = {
    debugFlag = false,
    blatherFlag = false,
    duelFlag = true,
    taxiFlag = true,
    gossipFlag = true,
    rewardFlag = false,
    repairFlag = true,
    bagFlag = true,
    bagWarn = 5,
    questLogFlag = true,
    questLogWarn = 5,
    vendJunkFlag = false,
    repairFlag = true
  }
}


--
-- Field handlers
--

-- DEBUG
function MultiTool:getDebugFlag(info)
  return self.db.profile.debugFlag
end
  
function MultiTool:setDebugFlag(info, value)
  self.db.profile.debugFlag = value
end

function MultiTool:getBlatherFlag(info)
  return self.db.profile.blatherFlag
end
  
function MultiTool:setBlatherFlag(info, value)
  self.db.profile.blatherFlag = value
end

-- DUEL
function MultiTool:getDuelFlag(info)
  return self.db.profile.duelFlag
end
  
function MultiTool:setDuelFlag(info, value)
  self.db.profile.duelFlag = value
end

-- TAXI
function MultiTool:getTaxiFlag(info)
  return self.db.profile.taxiFlag
end
  
function MultiTool:setTaxiFlag(info, value)
  self.db.profile.taxiFlag = value
end

-- QUEST GOSSIP
function MultiTool:getGossipFlag(info)
  return self.db.profile.gossipFlag
end
  
function MultiTool:setGossipFlag(info, value)
  self.db.profile.gossipFlag = value
end

-- REWARD CLONE
function MultiTool:getRewardFlag(info)
  return self.db.profile.rewardFlag
end
  
function MultiTool:setRewardFlag(info, value)
  self.db.profile.rewardFlag = value
end

-- REPAIR
function MultiTool:getRepairFlag(info)
  return self.db.profile.repairFlag
end
  
function MultiTool:setRepairFlag(info, value)
  self.db.profile.repairFlag = value
end

-- BAG MONITORING
function MultiTool:getBagFlag(info)
  return self.db.profile.bagFlag
end
  
function MultiTool:setBagFlag(info, value)
  self.db.profile.bagFlag = value
end

function MultiTool:getBagWarn(info)
  return self.db.profile.bagWarn
end
  
function MultiTool:setBagWarn(info, value)
  self.db.profile.bagWarn = value
end

-- QUEST LOG MONITORING
function MultiTool:getQuestLogFlag(info)
  return self.db.profile.questLogFlag
end
  
function MultiTool:setQuestLogFlag(info, value)
  self.db.profile.questLogFlag = value
end

function MultiTool:getQuestLogWarn(info)
  return self.db.profile.questLogWarn
end
  
function MultiTool:setQuestLogWarn(info, value)
  self.db.profile.questLogWarn = value
end

-- AUTO MERCHANT
function MultiTool:getRepairFlag(info)
  return self.db.profile.repairFlag
end
  
function MultiTool:setRepairFlag(info, value)
  self.db.profile.repairFlag = value
end

function MultiTool:getVendJunkFlag(info)
  return self.db.profile.vendJunkFlag
end
  
function MultiTool:setVendJunkFlag(info, value)
  self.db.profile.vendJunkFlag = value
end

--
-- Chat Command setup
-- 
function MultiTool:ChatCommand(input)
  if not input or input:trim() == "" then
    -- This would use the Ace config gui
    -- LibStub("AceConfigDialog-3.0"):Open("MultiTool")
    -- This uses the Bliz options pane instead
    InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
  else
    LibStub("AceConfigCmd-3.0").HandleCommand(MultiTool, L["SHORT_SLASH_CMD"], L["FULL_SLASH_CMD"], input)
  end
end


-- ------------------------------------------------
-- BASIC REQUIRED FUNCTIONS
-- ------------------------------------------------

-- Called when the addon is loaded
function MultiTool:OnInitialize()
  -- Using the AceDB for storage of our info
  self.db = LibStub("AceDB-3.0"):New("MultiToolDB", defaults, "Default")

  -- main options table stuff
  LibStub("AceConfig-3.0"):RegisterOptionsTable("MultiTool", options)
  -- adding to Blizzard options configuration
  self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("MultiTool", "MultiTool")
  -- registering the slash commands
  self:RegisterChatCommand(L["FULL_SLASH_CMD"], "ChatCommand")
  self:RegisterChatCommand(L["SHORT_SLASH_CMD"], "ChatCommand")
  self:RegisterChatCommand("help", "helpDialog")
  
  self:RegisterComm("MultiTool","OnCommReceived")
end


-- Called when the addon is enabled
function MultiTool:OnEnable()
  -- Register Duel requested event so we can swat em
  self:RegisterEvent("DUEL_REQUESTED", "duelRequested")
  -- Register merchant Show for autorepair and auto sell
	self:RegisterEvent("MERCHANT_SHOW", "autoMerchant")

--[[
	-- register ITEM_PUSH to fire off when item gained
  self:RegisterEvent("ITEM_PUSH", "updateBagSpace")

  -- COMMENTING OUT IN FAVOR OF ITEM_PUSH - I think it will be fired less
]]
	-- register bag_update to keep track of bag space
  self:RegisterEvent("BAG_UPDATE", "updateBagSpace")


  -- register quest log update so we can track that
  self:RegisterEvent("QUEST_LOG_UPDATE", "updateQuestLogSpace")
	
	-- Taxi hook
	hooksecurefunc("TakeTaxiNode",takeTaxiNodeHook)

  -- Quest gossip hooks
	hooksecurefunc("SelectGossipAvailableQuest", selectGossipAvailableQuestHook)
	hooksecurefunc("SelectGossipActiveQuest", selectGossipActiveQuestHook)
	hooksecurefunc("SelectGossipOption", selectGossipOptionHook)

	hooksecurefunc("SelectAvailableQuest", selectAvailableQuestHook)
	hooksecurefunc("SelectActiveQuest", selectActiveQuestHook)

	hooksecurefunc("AcceptQuest", acceptQuestHook)
	hooksecurefunc("CompleteQuest", completeQuestHook)
	hooksecurefunc("GetQuestReward", getQuestRewardHook)
	hooksecurefunc("DeclineQuest", declineQuestHook)

  -- update our bag count
  self:updateBagSpace()
  -- update our quest log count
  self:updateQuestLogSpace()

  self:Print("MultiTool "..myVersion.." ENABLED!")

end


-- Called when the addon is disabled
function MultiTool:OnDisable()
  --
end


-- ------------------------------------------------
-- EVENT HANDLERS
-- ------------------------------------------------

--
-- Auto Decline Duels
--
function MultiTool:duelRequested(challenger)
	if (self.db.profile.duelFlag) then
    if (self.db.profile.debugFlag) then
      self:Print("Duel denied from challenger:"..challenger)
    end -- debug
    CancelDuel();
  	StaticPopup_Hide("DUEL_REQUESTED");
	end -- if duelFlag
end -- function duelRequested()


--
-- Wrapper for auto repair and auto sell gray
--
function MultiTool:autoMerchant()
  -- let's sell first - just in case we need repair money
  self:autoSellJunk()
  -- Now, repair all
  self:autoRepair()
end


--
-- Auto Sell Junk - Inspired by CrapAway
--
function MultiTool:autoSellJunk()
  if (self.db.profile.vendJunkFlag) then
    self:debugMsg("autoSellJunk():: ENGAGED", "debug")

    -- iterate through bags
    for bag = 0, 4 do
      -- need to know the number of slots we have to look in
  		local numSlots = GetContainerNumSlots(bag)
      self:debugMsg("  bag"..tostring(bag)..": "..numSlots.." total slots", "blather")

      -- we have slots
  		if (numSlots > 0) then
        -- go through items in bag
        for slot = 1, numSlots + 1 do
          local currentItem = GetContainerItemLink(bag,slot)
          if (currentItem ~= nil) then
            -- Gotta get the item info using the link
            
            -- local currentItemQuality = Select(3, GetItemInfo(currentItem))
            local currentItemQuality = nil
             _,_,currentItemQuality,_,_,_,_,_,_,_ = GetItemInfo(currentItem)
            if (currentItemQuality ~= nil and currentItemQuality == 0) then
              UseContainerItem(bag,slot)
            end
          end
        end
  		end
  	end -- end for loop
  end -- if vendJunkFlag
end -- autoSellJunk()


--
-- Auto Repair All
--
function MultiTool:autoRepair()
	if (self.db.profile.repairFlag) then
    self:debugMsg("autoRepair():: ENGAGED", "debug")
    if (CanMerchantRepair()) then
			RepairAllItems()
		end
	end -- if repairFlag
end -- function autoRepair()


--
-- Bag total/free updater  (Based on BagSlots addon)
--
function MultiTool:updateBagSpace()
  -- to be efficient, only bother if we are flagged for it
	if (self.db.profile.bagFlag) then
    if (self.db.profile.debugFlag) then
      self:debugMsg("updateBagSpace():: ENGAGED", "debug")
      self:debugMsg("  bagWarn:"..tostring(self.db.profile.bagWarn), "blather")
      self:debugMsg("  totalBagSlots:"..tostring(totalBagSlots), "blather")
      self:debugMsg("  totalFreeBagSlots:"..tostring(totalFreeBagSlots), "blather")
      self:debugMsg("  totalUsedBagSlots:"..tostring(totalUsedBagSlots), "blather")
      self:debugMsg("  index:"..tostring(index), "blather")
    end -- debug

    local sendWarning = false
    local tmp_totalBagSlots = 0
    local tmp_totalFreeBagSlots = 0
    local tmp_totalUsedBagSlots = 0
   
    -- go through bags and add each to the totals
  	for bag = 0, 4 do
      self:debugMsg("Getting slots for bag "..tostring(bag), "blather")

  		local numSlots = GetContainerNumSlots(bag)

  		if (numSlots > 0) then
    		tmp_totalBagSlots = tmp_totalBagSlots + numSlots
  			local freeSlots = GetContainerNumFreeSlots(bag)
  			tmp_totalFreeBagSlots = tmp_totalFreeBagSlots + freeSlots
  			local usedSlots = numSlots - freeSlots
  			tmp_totalUsedBagSlots = tmp_totalUsedBagSlots + usedSlots

        if (self.db.profile.debugFlag) then
          self:debugMsg("Bag "..tostring(bag)..": has slots", "blather")
          self:debugMsg("  numSlots:"..tostring(numSlots), "blather")
          self:debugMsg("  tmp_totalBagSlots:"..tostring(tmp_totalBagSlots), "blather")
          self:debugMsg("  freeSlots:"..tostring(freeSlots), "blather")
          self:debugMsg("  tmp_totalFreeBagSlots:"..tostring(tmp_totalFreeBagSlots), "blather")
          self:debugMsg("  tmp_totalUsedBagSlots:"..tostring(tmp_totalUsedBagSlots), "blather")
        end
      else
        self:debugMsg("Bag "..tostring(bag)..": not present", "blather")
  		end
  	end -- end for loop
    
    self:debugMsg("Done counting bags", "debug")
    self:debugMsg("  tmp_totalFreeBagSlots:"..tostring(tmp_totalFreeBagSlots), "debug")
    self:debugMsg("  bagWarn:"..tostring(self.db.profile.bagWarn), "debug")

    -- check to see if this update is gain or loss of free slots
    if (tmp_totalFreeBagSlots < totalFreeBagSlots) then
      self:debugMsg("Free bag space went DOWN... checking agains thresholds", "debug")

      if (tmp_totalFreeBagSlots == 0) then
        self:debugMsg("  Bag Space EMPTY ... sendWarning = true", "debug")
        sendWarning = true      
      elseif (tmp_totalFreeBagSlots == self.db.profile.bagWarn) then
        self:debugMsg("  Threshold met ... sendWarning = true", "debug")
        sendWarning = true
      else
        self:debugMsg("  Threshold NOT met... sendWarning = false", "debug")
      end
    end
    
    self:debugMsg("Handling Warnings", "debug")    
    -- Send the warning message if we have one
    if (sendWarning) then 
      self:debugMsg("  sendWarning is true: sending warning", "debug")

      local prefix = "MultiTool"
      local msg = {
        op = "BAG_WARN",
        free_slots = tmp_totalFreeBagSlots
      }
      local text = self:Serialize(msg)
      local distribution = "RAID"
      local target = UnitName("player")
      local prio = "ALERT"

      -- rpcCommandSender adds checking of the noComm value before sending
      self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
    else 
      self:debugMsg("  sendWarning is false: not sending comms", "blather")
    end -- if sendWarning

    self:debugMsg("Updating master bag slot count values", "debug")
    -- Update the addon's copy of the bag levels
    totalBagSlots = tmp_totalBagSlots     
    totalFreeBagSlots = tmp_totalFreeBagSlots    
    totalUsedBagSlots = tmp_totalUsedBagSlots
    
    self:debugMsg("  totalBagSlots:"..tostring(totalBagSlots), "blather")
    self:debugMsg("  totalFreeBagSlots:"..tostring(totalFreeBagSlots), "blather")
    self:debugMsg("  totalUsedBagSlots:"..tostring(totalUsedBagSlots), "blather")
    
	end -- end if bagFlag
end -- end of funcction


--
-- Quest Log total/free updater
--
function MultiTool:updateQuestLogSpace()
  -- to be efficient, only bother if we are flagged for it
	if (self.db.profile.questLogFlag) then
    self:debugMsg("updateQuestLogSpace():: ENGAGED", "debug")
    
    -- if they ever up the num of quests available...
    local constant_questLogSize = 25
    
    local sendWarning = false
    local tmp_totalQuestLogEntries = 0
    local tmp_totalFreeQuestLogSlots = 0

    tmp_totalQuestLogEntries = GetNumQuestLogEntries()
    tmp_totalFreeQuestLogSlots = constant_questLogSize - tmp_totalQuestLogEntries

    -- check to see if this update is gain or loss of free slots
    if (tmp_totalFreeQuestLogSlots < totalFreeQuestLogSlots) then
      self:debugMsg("Free quest log slots DOWN... checking agains thresholds", "debug")

      if (tmp_totalFreeQuestLogSlots == 0) then
        self:debugMsg("  Quest Log FULL ... sendWarning = true", "debug")
        sendWarning = true      
      elseif (tmp_totalFreeQuestLogSlots == self.db.profile.questLogWarn) then
        self:debugMsg("  Threshold met ... sendWarning = true", "debug")
        sendWarning = true
      else
        self:debugMsg("  Threshold NOT met... sendWarning = false", "debug")
      end
    end
    
    -- Send the warning message if we have one
    if (sendWarning) then 
      local prefix = "MultiTool"
      local msg = {
        op = "QUEST_LOG_WARN",
        free_space = tmp_totalFreeQuestLogSlots
      }
      local text = self:Serialize(msg)
      local distribution = "RAID"
      local target = UnitName("player")
      local prio = "NORMAL"

      -- rpcCommandSender adds checking of the noComm value before sending
      self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
    end -- if sendWarning
    
    -- last thing:: Update the addon's copy of the freeSlot levels
    totalFreeQuestLogSlots = tmp_totalFreeQuestLogSlots    
	else
    self:debugMsg("questLogFlag is false... updateQuestLogSpace() skipped", "debug")
	end -- end if questLogFlag
end -- end of funcction

-- ------------------------------------------------
-- HOOKED FUNCTIONS
-- ------------------------------------------------

--
-- Quest Functions
--
function selectGossipActiveQuestHook(index, noComm)
  local self = MultiTool
  
  if (self.db.profile.gossipFlag) then
    self:debugMsg("selectGossipActiveQuestHook():: ENGAGED", "debug")
    self:debugMsg("  index:"..tostring(index), "blather")
    self:debugMsg("  noComm:"..tostring(noComm), "blather")
    
    local prefix = "MultiTool"
    local msg = {
      op = "QUEST_GOSSIP",
      sub_op = "SELECT_GOSSIP_ACTIVE",
      index = index
    }
    local text = self:Serialize(msg)
    
    local distribution = "RAID"
    local target = UnitName("player")
    local prio = "ALERT"
    
    -- rpcCommandSender adds checking of the noComm value before sending
    self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
  else
    self:debugMsg("gossipFlag is false... not sending comms", "debug")
  end
end


function getQuestRewardHook(index, noComm)
  local self = MultiTool
  
  if (self.db.profile.gossipFlag) then
    self:debugMsg("getQuestRewardHook():: ENGAGED", "debug")
    self:debugMsg("  index:"..tostring(index), "blather")
    self:debugMsg("  noComm:"..tostring(noComm), "blather")

    local prefix = "MultiTool"
    local msg = {
      op = "QUEST_GOSSIP",
      sub_op = "GET_QUEST_REWARD",
      index = index
    }
    local text = self:Serialize(msg)
    
    local distribution = "RAID"
    local target = UnitName("player")
    local prio = "ALERT"
    
    -- rpcCommandSender adds checking of the noComm value before sending
    self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
  else
    self:debugMsg("gossipFlag is false... not sending comms", "debug")
  end
end


function selectAvailableQuestHook(index, noComm)
  local self = MultiTool
  
  if (self.db.profile.gossipFlag) then
    self:debugMsg("selectAvailableQuestHook():: ENGAGED", "debug")
    self:debugMsg("  index:"..tostring(index), "blather")
    self:debugMsg("  noComm:"..tostring(noComm), "blather")

    local prefix = "MultiTool"
    local msg = {
      op = "QUEST_GOSSIP",
      sub_op = "SELECT_AVAILABLE",
      index = index
    }
    local text = self:Serialize(msg)
    
    local distribution = "RAID"
    local target = UnitName("player")
    local prio = "ALERT"
    
    -- rpcCommandSender adds checking of the noComm value before sending
    self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
  else
    self:debugMsg("gossipFlag is false... not sending comms", "debug")
  end
end


function selectGossipAvailableQuestHook(index, noComm)
  local self = MultiTool
  
  if (self.db.profile.gossipFlag) then
    self:debugMsg("selectGossipAvailableQuestHook():: ENGAGED", "debug")
    self:debugMsg("  index:"..tostring(index), "blather")
    self:debugMsg("  noComm:"..tostring(noComm), "blather")

    local prefix = "MultiTool"
    local msg = {
      op = "QUEST_GOSSIP",
      sub_op = "SELECT_GOSSIP_AVAILABLE",
      index = index
    }
    local text = self:Serialize(msg)
    
    local distribution = "RAID"
    local target = UnitName("player")
    local prio = "ALERT"
    
    -- rpcCommandSender adds checking of the noComm value before sending
    self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
  else
    self:debugMsg("gossipFlag is false... not sending comms", "debug")
  end
end


function selectGossipOptionHook(index, noComm)
  local self = MultiTool

  if (self.db.profile.gossipFlag) then
    self:debugMsg("selectGossipOptionHook():: ENGAGED", "debug")
    self:debugMsg("  index:"..tostring(index), "blather")
    self:debugMsg("  noComm:"..tostring(noComm), "blather")

    local prefix = "MultiTool"
    local msg = {
      op = "QUEST_GOSSIP",
      sub_op = "SELECT_OPTION",
      index = index
    }
    local text = self:Serialize(msg)
    
    local distribution = "RAID"
    local target = UnitName("player")
    local prio = "ALERT"
    
    -- rpcCommandSender adds checking of the noComm value before sending
    self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
  else
    self:debugMsg("gossipFlag is false... not sending comms", "debug")
  end
end


function selectActiveQuestHook(index, noComm)
  local self = MultiTool

  if (self.db.profile.gossipFlag) then
    self:debugMsg("selectActiveQuestHook():: ENGAGED", "debug")
    self:debugMsg("  index:"..tostring(index), "blather")
    self:debugMsg("  noComm:"..tostring(noComm), "blather")
    
    local prefix = "MultiTool"
    local msg = {
      op = "QUEST_GOSSIP",
      sub_op = "SELECT_ACTIVE",
      index = index
    }
    local text = self:Serialize(msg)
    
    local distribution = "RAID"
    local target = UnitName("player")
    local prio = "ALERT"
    
    -- rpcCommandSender adds checking of the noComm value before sending
    self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
  else
    self:debugMsg("gossipFlag is false... not sending comms", "debug")
  end
end


function acceptQuestHook(noComm)
  local self = MultiTool
  
  if (self.db.profile.gossipFlag) then
    self:debugMsg("acceptQuestHook():: ENGAGED", "debug")
    self:debugMsg("  noComm:"..tostring(noComm), "blather")
    
    local prefix = "MultiTool"
    local msg = {
      op = "QUEST_GOSSIP",
      sub_op = "ACCEPT_QUEST"
    }
    local text = self:Serialize(msg)
    
    local distribution = "RAID"
    local target = UnitName("player")
    local prio = "ALERT"
    
    -- rpcCommandSender adds checking of the noComm value before sending
    self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
  else
    self:debugMsg("gossipFlag is false... not sending comms", "debug")
  end
end


function completeQuestHook(noComm)
  local self = MultiTool
  
  if (self.db.profile.gossipFlag) then
    self:debugMsg("completeQuestHook():: ENGAGED", "debug")
    self:debugMsg("  noComm:"..tostring(noComm), "blather")
    
    local prefix = "MultiTool"
    local msg = {
      op = "QUEST_GOSSIP",
      sub_op = "COMPLETE_QUEST"
    }
    local text = self:Serialize(msg)
    
    local distribution = "RAID"
    local target = UnitName("player")
    local prio = "ALERT"
    
    -- rpcCommandSender adds checking of the noComm value before sending
    self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
  else
    self:debugMsg("gossipFlag is false... not sending comms", "debug")
  end
end


function declineQuestHook(noComm)
  local self = MultiTool
  
  if (self.db.profile.gossipFlag) then
    self:debugMsg("declineQuestHook():: ENGAGED", "debug")
    self:debugMsg("  noComm:"..tostring(noComm), "blather")
    
    local prefix = "MultiTool"
    local msg = {
      op = "QUEST_GOSSIP",
      sub_op = "DECLINE_QUEST"
    }
    local text = self:Serialize(msg)
    
    local distribution = "RAID"
    local target = UnitName("player")
    local prio = "ALERT"
    
    -- rpcCommandSender adds checking of the noComm value before sending
    self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
  else
    self:debugMsg("gossipFlag is false... not sending comms", "debug")
  end
end


--
-- Taxi following
--
function takeTaxiNodeHook(index, noComm) 
  local self = MultiTool
    
  -- Only use the additional functionality IF taxiFlag is true
  if (self.db.profile.taxiFlag) then
    self:debugMsg("takeTaxiNodeHook():: ENGAGED", "debug")
    self:debugMsg("  index:"..tostring(index), "blather")
    self:debugMsg("  noComm:"..tostring(noComm), "blather")
    
    local prefix = "MultiTool"
    local msg = {
      op = "TAKETAXI",
      node = index,
      nodeName = TaxiNodeName(index)
    }
    local text = self:Serialize(msg)

    local distribution = "RAID"
    local target = UnitName("player")
    local prio = "ALERT"   
    
    -- rpcCommandSender adds checking of the noComm value before sending
    self:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
  else
    self:debugMsg("taxiFlag is false... not sending comms", "debug")
  end
end -- end function takeTaxiNodeHook


-- ------------------------------------------------
-- COMM-BASED FUNCTIONS
-- ------------------------------------------------

--
-- Handles received commands
--
function MultiTool:OnCommReceived(prefix, message, distribution, target)
  -- Deserialize incoming message
  local success, o = self:Deserialize(message)
  if (self.db.profile.debugFlag) then
    self:debugMsg("OnCommReceived():: ENGAGED", "debug")
    self:debugMsg("  prefix:"..tostring(prefix), "blather")
    self:debugMsg("  message:"..tostring(message), "blather")
    self:debugMsg("  distribution:"..tostring(distribution), "blather")
    self:debugMsg("  target:"..tostring(target), "blather")
    self:debugMsg("  player:"..UnitName("player"), "blather")
    self:debugMsg("  Deserialize success:"..tostring(success), "blather")
    self:debugMsg("  Message Content:"..tostring(o), "blather")
  end -- end if debugFlag

  -- don't do anyting if we could not deserialize
  if (success == false) then
    self:debugMsg("Failed to deserialize the message - OnCommReceived can not process any further", "error")
    return -- Failure
  else
    --
    -- TEMPLATE - Not actually used
    --
    if (o["op"] == "foo") then
      if (self.db.profile.actionFlag) then
        self:debugMsg("FOO op ENGAGED", "debug")

        -- Put any extra conditions here
        if ("someCondition" == "some value") then
          self:debugMsg("debug_text_here", "blather")
          return -- do nothing
        else
          -- do the action
          self:debugMsg("debug_text_here", "blather")

          -- actual command such as TakeTaxiNode(o["node"], "noComm")
        end -- ending extra conditions
      end -- if gossipFlag
    end -- end if op == foo
    
    --
    -- Bag Space Warning
    --
    if (o["op"] == "BAG_WARN") then
      if (self.db.profile.bagFlag) then
        self:debugMsg("BAG_WARN op engaged", "debug")
        
        -- Only do the warning message IF it's empty bag message or if it's AT the threshold
        if (o["free_slots"] == 0 or o["free_slots"] == self.db.profile.bagWarn) then
          self:debugMsg("  conditions met - displaying warning", "blather")
          
          -- need to localize this
          local tmp_msgText = string.format(L["BAG_WARN_MSG"], target, o["free_slots"])
          self:debugMsg("  tmp_msgText: "..tmp_msgText, "blather")

          -- DO THE WARNING
          UIErrorsFrame:AddMessage(tmp_msgText, 1.0, 1.0, 1.0, 5.0)
        else
          self:debugMsg("  conditions NOT met - QUEST_WARN not sending msg", "blather")
        end
      end -- end of flagCheck
    end -- end if op == BAG_WARN

    --
    -- Quest Log Warning
    --
    if (o["op"] == "QUEST_LOG_WARN") then
      if (self.db.profile.questLogFlag) then
        self:debugMsg("QUEST_WARN op engaged", "debug")

        -- Only do the warning message IF it's empty bag message or if it's AT the threshold
        if (o["free_slots"] == 0 or o["free_slots"] == self.db.profile.questLogWarn) then
          self:debugMsg("  conditions met - displaying warning", "blather")

          local tmp_msgText = string.format(L["QUEST_LOG_WARN_MSG"], target, o["free_slots"])
          
          -- DO THE WARNING
          UIErrorsFrame:AddMessage(self.tmp_msgText, 1.0, 1.0, 1.0, 5.0)
        else
          self:debugMsg("  conditions NOT met - QUEST_WARN not sending msg", "blather")
        end
      end -- end of flagCheck
    end -- end if op == BAG_WARN

    --
    -- Quest Gossip handler for cloning actions at quest givers
    --
    if (o["op"] == "QUEST_GOSSIP") then
      if (self.db.profile.gossipFlag) then
        -- Put global quest gossip checks here
        if ("someCondition" == "some bad value") then
          -- conditions mean we DON'T do anything
          self:debugMsg("Global QUEST_GOSSIP global condition failed - not doing anything", "blather")

          return -- do nothing
        else
          -- do the action
          self:debugMsg("QUEST_GOSSIP global conditions succeded - parsing subcommands", "blather")
                   
          if (o["sub_op"] == "SELECT_GOSSIP_AVAILABLE") then
            -- tacking on noComm so we don't cause message storm
         		SelectGossipAvailableQuest(o["index"], "noComm")
          elseif (o["sub_op"] == "SELECT_AVAILABLE") then
            -- tacking on noComm so we don't cause message storm
         		SelectAvailableQuest(o["index"], "noComm")
          elseif (o["sub_op"] == "SELECT_OPTION") then
            -- tacking on noComm so we don't cause message storm
		        SelectGossipOption(o["index"], "noComm");
          elseif (o["sub_op"] == "SELECT_ACTIVE") then
            -- tacking on noComm so we don't cause message storm
		        SelectActiveQuest(o["index"], "noComm");
          elseif (o["sub_op"] == "ACCEPT_QUEST") then
            -- tacking on noComm so we don't cause message storm
		        AcceptQuest("noComm")
          elseif (o["sub_op"] == "GET_QUEST_REWARD") then
            -- This one's special.. using logic from QuestGuru to make sure we
            -- don't allow cloning of reward choices
            if (self.db.profile.rewardFlag or GetNumQuestChoices() < 2) then
              self:debugMsg(tostring(GetNumQuestChoices()).." choice - ok to run", "debug")
              -- tacking on noComm so we don't cause message storm
    		      GetQuestReward(o["index"],"noComm")	  
            else
              self:debugMsg(tostring(GetNumQuestChoices()).." choices - Blocking getQuestReward", "debug")
            end
          elseif (o["sub_op"] == "SELECT_GOSSIP_ACTIVE") then
            -- tacking on noComm so we don't cause message storm
		        SelectGossipActiveQuest(o["index"], "noComm")
          elseif (o["sub_op"] == "COMPLETE_QUEST") then
            -- tacking on noComm so we don't cause message storm
		        CompleteQuest("noComm")
          elseif (o["sub_op"] == "DECLINE_QUEST") then
            -- tacking on noComm so we don't cause message storm
		        DeclineQuest("noComm")
          end


        end -- ending extra conditions
      end -- if gossipFlag
    end -- end if op == foo
    
    --
    -- TAKETAXI
    --
    if (o["op"] == "TAKETAXI") then

      -- Only continue IF the taxiFlag is true
      if (self.db.profile.taxiFlag) then

        local numNodes = NumTaxiNodes()
        self:debugMsg("  numNodes: "..tostring(numNodes), "blather")
        
        -- see if we're at a flightmaster
        if (numNodes > 0) then
  
          -- Fix for ticket #1
          if (o["nodeName"] ~= "INVALID" and o["nodeName"] ~= TaxiNodeName(o["node"])) then
            self:debugMsg("TAKETAXI: Passed Index to NodeName mismatch... searching by name", "debug")
            self:debugMsg("  o.node: "..tostring(o["node"]), "debug")
            self:debugMsg("  o.nodeName: "..tostring(o["nodeName"]), "debug")
            self:debugMsg("  local nodeName: "..tostring(TaxiNodeName(o["node"])), "debug")
            
            --[[
            -- for one of a number of reasons, the requested node's name does not 
            -- match the node name of the index we were passed. We will try to do 
            -- a more intensive lookup and if we find the issue, we will replace 
            -- the passed index with a local index value
            ]]
            
            local matchFoundFlag = false
            self:debugMsg("  numNodes: "..tostring(numNodes), "debug")
  
            self:debugMsg("Iterating through available nodes", "debug")
            local currentNode = 1
            local indexChanged = false
  
            --[[
            -- while (matchFoundFlag == false or currentNode < (numNodes +1) or currentNode < 50) do
            ]]
            for currentNode = 1, numNodes, 1 do
              local currentNodeName = TaxiNodeName(currentNode)
               self:debugMsg("  currentNode: "..tostring(currentNode), "blather")
               self:debugMsg("  currentNodeName: "..tostring(currentNodeName), "blather")
  
              if (currentNodeName == tostring(o["nodeName"])) then
                self:debugMsg("MATCH FOUND... setting new node index: "..tostring(currentNode), "blather")
                o["node"] = currentNode
                indexChanged = true
                break
              end -- if currentnodeName == o.nodeName
              currentNode = currentNode + 1
            end -- for loop
            
            if (indexChanged == false) then
              --failure
              msgString = string.format(L["TAXI_CANT_FOLLOW"], o["nodeName"])
              self:debugMsg(msgString, "warn")
              -- force invalid node so that we don't try to take one anyway
              o["node"] = 0
              self:debugMsg("setting o.node=0", "debug")
            end
          end -- nodename ~= INVALID and nodename ~TaxiNodeName(o.node)

          if (o["nodeName"] ~= TaxiNodeName(o["node"])
              or UnitOnTaxi("player") 
              or o["nodeName"] == "INVALID" 
              or target == UnitName("player")
          ) 
          then
            -- NoSend
            if (self.db.profile.debugFlag) then
              self:debugMsg("TAKETAXI op NOT engaged", "debug")
              self:debugMsg("  searchFailed "..tostring(searchFailed), "blather")
              self:debugMsg("  Am I on taxi? "..tostring(UnitOnTaxi("player")), "blather")
              self:debugMsg("  actual node requested: "..tostring(o["node"]).." ("..o["nodeName"]..")", "blather")
              self:debugMsg("  effective node requested: "..tostring(o["node"]).." ("..TaxiNodeName(o["node"])..")", "blather")
            end
            return -- DON'T DO ANYTHING: this stops a message storm
          else
            self:debugMsg("TAKETAXI op engaged", "debug")
            -- by passing the extra arg "noComm" to the takeTaxiNode, I think I solve the msg storms
            TakeTaxiNode(o["node"], "noComm")
          end -- big multiline block
        else
          -- taxiFlag set, but NOT at flightmaster
          self:debugMsg(L["TAXI_NO_FLIGHTMASTER"], "notice")
        end -- if numNodes
      else
        self:debugMsg("taxiFlag is false... ignoring TAKETAXI request", "debug")
      end -- end if taxiFlag
    end -- end if op==TAKETAXI
  end -- if success==false
end -- function 


--
-- Wraps up Remote procedure call commands with a "nocomm" throttler
--
function MultiTool:rpcCommandSender(prefix, text, distribution, target, prio, noComm)
  --debug
  if (self.db.profile.debugFlag) then
    self:debugMsg("rpcCommandSender():: engaged", "debug")
    self:debugMsg("  prefix: "..tostring(prefix), "blather")
    self:debugMsg("  text: "..tostring(text), "blather")
    self:debugMsg("  distribution: "..tostring(distribution), "blather")
    self:debugMsg("  target: "..tostring(target), "blather")
    self:debugMsg("  prio: "..tostring(prio), "blather")
    self:debugMsg("  noComm flag: "..tostring(noComm), "blather")
  end

  -- New throttling - DO NOT send msg unless noComm is nil
  if (noComm) then
    self:debugMsg("noComm flag IS set, NOT sending command", "blather")
    -- Teh Gogglez - They do NOTHING!
    return
  else
    self:debugMsg("noComm flag NOT set, sending command", "blather")
    if (GetNumPartyMembers() == 0 and (distribution == "RAID" or distribution == "PARTY")) then
      -- NOT in party, but trying to send to party - modify distirbution and target
      distribution = "WHISPER"
      target = GetUnitName("player")
    end
    self:SendCommMessage(prefix, text, distribution, target, prio)     
  end
end

-- ------------------------------------------------
-- UTILITY FUNCTIONS
-- ------------------------------------------------
function MultiTool:helpCommand()
  self:Print("MultTool Help")
  self:Print("")
  self:Print("COMMANDS:")
  self:Print("  /MultiTool help - Show this help menu")
  self:Print("")
end

-- Making debugging a little simpler
function MultiTool:debugMsg(text, level)
  if (level == nil) then
    return -- do nothing
  elseif (level == "error") then
    UIErrorsFrame:AddMessage(text, 1.0, 1.0, 1.0, 5.0)
  elseif (level == "warn") then
    UIErrorsFrame:AddMessage(text, 1.0, 1.0, 1.0, 5.0)
  elseif (level == "notice") then
    self:Print(level..": "..text)
  elseif (level == "debug" and (self.db.profile.debugFlag or self.db.profile.blatherFlag)) then
    self:Print(level..": "..text)
  elseif (level == "blather" and self.db.profile.blatherFlag) then
    self:Print(level..": "..text)
  end
end
